;;;; "logic.scm" utility functions for synchronous simulation.	-*-scheme-*-
;;; Copyright (C) 1995, 1996, 1997, 1998 Aubrey Jaffer

;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

(define header
  "@(#) $Id: logic.scm,v 1.29 1998/11/09 02:43:31 jaffer Exp $")

(require 'models)			;(in-vicinity (program-vicinity) "models.scm")

(if (not (defmacro? 'casev))
(defmacro casev (key . clauses)
  (let ((clauses
	 (map (lambda (clause)
		(if (list? (car clause))
		    (cons (apply
			   append
			   (map (lambda (elt)
				  (case elt
				    ((unquote) '(unquote))
				    ((unquote-splicing) '(unquote-splicing))
				    (else
				     (eval (list 'quasiquote (list elt))))))
				(car clause)))
			  (cdr clause))
		    clause))
	      clauses)))
    `(case ,key ,@clauses))))

(define (xor a b)
  (cond ((and (boolean? a) (boolean? b))
	 (not (eq? a b)))
	(else (synch:warn 'xor "requires boolean arguments" a b)
	      (or (and (not a) b) (and a (not b))))))

;;;; Bit-reverse taken from lcdecode.scm
(define (bit-reverse width in)
  (cond ((zero? width) 0)
	(else
	 (+ (bit-reverse (+ -1 width) (ash in -1))
	    (ash (logand 1 in) (+ -1 width))))))

(require 'random)
(define (random-% %) (< (random 100) %))

;; Non-booleans set bits left shifted by 8 bits.
(require 'common-list-functions)
(define (booleans-to-number . bools)
  (cond ((every boolean? bools)
	 (do ((bs bools (cdr bs))
	      (acc 0 (+ acc acc (if (car bs) 1 0))))
	     ((null? bs) acc)))
	(else
	 ;;(synch:warn 'booleans-to-number 'non 'boolean 'in bools)
	 (do ((bs bools (cdr bs))
	      (acc 0 (+ acc acc (case (car bs)
				  ((#t) 1)
				  ((#f) 0)
				  (else 256)))))
	     ((null? bs) (logior acc #x-10000))))))

(define (count . booleans)
  (do ((booleans booleans (cdr booleans))
       (count 0 (+ count (cond ((not (car booleans)) 0)
			       ((boolean? (car booleans)) 1)
			       (else (synch:warn "COUNTing non-boolean?"
						 booleans)
				     1)))))
      ((null? booleans) count)))

(define (with-set-bit n b idx)
  (if b
      (logior n (ash 1 idx))
      (logand n (lognot (ash 1 idx)))))

(define (bit-subst orig start end new)
  (define mask (ash (+ -1 (ash 1 (+ 1 (- end start)))) start)
    (logior (logand mask (ash new start))
	    (logand (lognot maskcomp) b))))

(define (binary->gray n)
  (logxor n (ash n -1)))

(define (even-parity? n)
  (even? (logcount n)))

(define (gray->binary n)
  (if (negative? n)
      (error 'gray->binary 'negative? n)
      (do ((ntmp n (ash ntmp -1))
	   (nbits 0 (+ nbits 1)))
	  ((zero? ntmp)
	   (do ((i (- nbits 1) (- i 1))
		(term (- (ash 1 nbits) 1) (ash term -1))
		(sum 0 (if (logbit? i n) (logxor sum term) sum)))
	       ((negative? i) sum))))))

;; UP? must be boolean, #t means increment, #f means decrement.
(define (gray:inc up? n)
  (cond ((zero? n) (if up? 1 (error 'gray:decrement 'zero?)))
	((negative? n) (error 'gray:inc 'negative? n))
	(else
	 (do ((ntmp n (ash ntmp -1))
	      (i-set 0 (+ i-set 1)))
	     ((odd? ntmp)
	      (do ((ntmp (ash ntmp -1) (ash ntmp -1))
		   (even-parity? #f (if (odd? ntmp)
					(not even-parity?)
					even-parity?)))
		  ((zero? ntmp)
		   (if (eq? up? even-parity?)
		       (logxor n 1)
		       (logxor n (ash 1 (+ 1 i-set)))))))))))

(define (gray:increment n)
  (gray:inc #t n))
(define (gray:decrement n)
  (gray:inc #f n))

;; Return 1 if N1 is after N2 in Gray code sequence, -1 if
;; before, 0 if equal.
(define (gray-signum n1 n2)
  (do ((nbits 0 (+ nbits 1))
       (dif (logxor n1 n2) (ash dif -1)))
      ((zero? dif)
       (if (zero? nbits) 0
	   (let loop ((ntmp (ash n1 (- nbits)))
		      (res (if (logbit? (- nbits 1) n1) 1 -1)))
	     (if (zero? ntmp) res
		 (loop (ash ntmp -1)
		       (if (odd? ntmp) (- res) res))))))))

(define (gray:<? n1 n2)
  (negative? (gray-signum n1 n2)))
(define (gray:<=? n1 n2)
  (not (positive? (gray-signum n1 n2))))
(define (gray:>? n1 n2)
  (positive? (gray-signum n1 n2)))
(define (gray:>=? n1 n2)
  (not (negative? (gray-signum n1 n2))))

(define (constant? x)
  (or (number? x) (boolean? x)))
(define (constants? . args)
  (every constant? args))

(define (compose-sets array . functions)
  (do ((array array ((car functions) array))
       (functions functions (cdr functions)))
      ((null? functions) array)))

(define *deferred-sets* '())

(if (not (defined? logaref))
    (define logaref
      (let ((orig-array-ref array-ref))
	(lambda args
	  (cond ((= (+ -1 (length args)) (+ 1 (array-rank (car args))))
		 (let ((idxlist (reverse (cdr (reverse args)))))
		   (logbit? (car (last-pair args))
			    (apply orig-array-ref idxlist))))
		(else			;(= idxs rank)
		 (apply orig-array-ref args)))))))
(if (not (defined? logaset!))
    (define (logaset! args)
      (cond ((= (+ -2 (length args)) (+ 1 (array-rank (car args))))
	     (let ((idxlist (reverse (cdr (reverse (cddr args))))))
	       (apply array-set!
		      (car args)
		      (with-set-bit
		       (apply array-ref (car args) idxlist)
		       (cadr args)
		       (car (last-pair args)))
		      idxlist)))
	    (else			;(= idxs rank)
	     (apply array-set! args)))))

(define (array-set . args)
  (set! *deferred-sets* (cons args *deferred-sets*))
  (car args))

(define (update-deferred-sets!)
  (for-each (lambda (args)
	      (apply logaset! args))
	    (reverse *deferred-sets*))
  (set! *deferred-sets* '()))

;;;; Here is an unpleasant little routine:
;;;  Defaults strongly (and often) to 0.
(define (number-or name . args)
  (define ans #f)
  (define ind 0)
  (cond ((not (symbol? name))
	 (synch:error "number-or's first argument not a symbol:" name)))
  (for-each (lambda (a)
	      (set! ind (+ 1 ind))
	      (cond ((not a))
		    ((eqv? a ans) (synch:warn name ind "redundant term" a))
		    (ans (synch:warn name ind "number-oring" a ans)
			 (set! ans (logior a ans)))
		    (else (set! ans a))))
	    args)
  (or ans 0))

(define (number-check name a1)
  (cond ((not (number? a1))
	 (synch:warn name "not a number" a1))
	((negative? a1)
	 (synch:warn name "unstable argument" a1)))
  #t)

(define (check= name x y)
  (cond ((not (eqv? x y))
	 (synch:warn name "unequal" x y)
	 ;;(synch:chat "<>")
	 ))
  y)

(define (sign x) (cond ((negative? x) -1)
		       ((positive? x) 1)
		       (else #f)))

(define (box-name sym)
  (let* ((str (symbol->string sym))
	 (len (string-length str))
	 (alpha-end
	  (do ((i (+ -1 len) (+ -1 i)))
	      ((or (negative? i) (not (char-numeric? (string-ref str i))))
	       (+ 1 i)))))
    (if (eqv? len alpha-end)
	sym
	(string->symbol (string-append (substring str 0 alpha-end) "["
				       (substring str alpha-end len) "]")))))

(define (vector-bounds sym)
  (let* ((str (symbol->string sym))
	 (len (string-length str))
	 (bound1-start
	  (do ((i (+ -1 len) (+ -1 i)))
	      ((or (negative? i) (char=? #\[ (string-ref str i)))
	       (if (negative? i) len (+ 1 i)))))
	 (bound1-end
	  (do ((i bound1-start (+ i 1)))
	      ((or (>= i len)
		   (memv (string-ref str i) '(#\. #\:)))
	       i)))
	 (bound2-start (+ (if (and (< bound1-end len)
				   (char=? #\: (string-ref str bound1-end)))
			      1 2)
			  bound1-end))
	 (bound2-end
	  (do ((i bound2-start (+ i 1)))
	      ((or (>= i len) (char=? #\] (string-ref str i)))
	       i)))
	 (bound1 (and (not (=? bound1-start bound1-end))
		      (substring str bound1-start bound1-end)))
	 (bound2 (and (not (=? bound2-start bound2-end))
		      (substring str bound2-start bound2-end))))
    (and bound1 bound2 (list (string->symbol
			      (substring str 0 (+ -1 bound1-start)))
			     (string->number bound1)
			     (string->number bound2)))))

(define (array-bounds sym)
  (let ((ini (signal-table:get-init-fun *block* sym)))
    ;;(print 'array-bounds sym '-> ini)
    (cond ((symbol? ini) (set! ini (slib:eval ini))))
    (cond ((array? ini) (array-shape ini) '())
	  ((not (pair? ini)) '())
	  (else
	   (case (car ini)
	     ((make-ram)
	      `(0 (- ,(cadr ini) 1)))
	     ((make-fifo)
	      `(-2 (- ,(cadr ini) 1)))
	     ((make-ram-array)
	      (cddr ini))
	     (else '()))))))

(define (vectorized? sym)
  ;;(set! sym (symbol->string sym))
  ;;(char=? #\] (string-ref sym (+ -1 (string-length sym))))
  (vector-bounds sym)
  )

(define (devectorize sym ind)
  (let* ((bounds (vector-bounds sym)))
    (cond (bounds
	   (let* ((name (car bounds))
		  (lower (cadr bounds))
		  (upper (caddr bounds)))
	     (cond ((> lower upper) (set! lower (caddr bounds))
				    (set! upper (cadr bounds))))
	     (cond ((>= upper ind lower) name)
		   (else ;;(report "devectorize: index out of bounds " ind sym) name
			 #f))))
	  (else (report "devectorize: index of scalar " ind sym)
		sym))))

(define (revectorize sep sym ind1 ind2)
  (let* ((bounds (vector-bounds sym)))
    (cond (bounds
	   (let* ((name (car bounds))
		  (lower (cadr bounds))
		  (upper (caddr bounds)))
	     (cond ((>= upper ind2 ind1 lower)
		    (string->symbol
		     (string-append
		      (symbol->string name)
		      "[" (number->string ind1)
		      sep (number->string ind2) "]")))
		   ((>= lower ind2 ind1 upper)
		    (string->symbol
		     (string-append
		      (symbol->string name)
		      "[" (number->string ind2)
		      sep (number->string ind1) "]")))
		   (else (report "revectorize: index out of bounds "
				 ind1 ind2 sym)
			 name))))
	  (else (report "revectorize: index of scalar " ind sym) sym))))

(define (refvectorize sym ind)
  (let* ((bounds (vector-bounds sym)))
    (cond ((not bounds) (report "refvectorize: index of scalar " ind sym) sym)
	  (else
	   (let* ((first (cadr bounds))
		  (last (caddr bounds))
		  (range (- first last))
		  (incer (sign range)))
	     (cond ((not (or (<= first ind last)
			     (<= last ind first)))
		    (report "refvectorize: index out of bounds " ind sym)
		    sym)
		   (else
		    (string->symbol
		     (string-append (symbol->string (car bounds))
				    "[" (number->string ind) "]")))))))))

;;; SCM bug workaround
(cond ((not (= 92 (bit-extract #x5c5c5c5c 0 8)))
       (dprintf "Isn't bit-extract fixed yet?\n")
       (set! bit-extract
	     (lambda (n start end)
	       (logand (- (integer-expt 2 (- end start)) 1)
		       (ash n (- start)))))))

;;;			      -=-=-=-=-

(define (symbol-append . args)
  (string->symbol (apply string-append
			 (map
			  (lambda (x)
			    (cond ((string? x) x)
				  ((number? x) (number->string x))
				  ((symbol? x) (symbol->string x))
				  ((not x) "")
				  (else (slib:error
					 'wrong-type-to 'symbol-append x))))
			  args))))

(define (ptag-prefix name)
  (and (symbol? name)
       (let* ((str (symbol->string name))
	      (len (string-length str)))
	 (do ((i 0 (+ 1 i)))
	     ((or (>= i len) (char=? #\: (string-ref str i)))
	      (if (>= i len)
		  #f
		  (string->symbol (substring str 0 i))))))))

(define (ptag-stripped name)
  (if (symbol? name)
      (let* ((str (symbol->string name))
	     (len (string-length str)))
	(do ((i 0 (+ 1 i)))
	    ((or (>= i len) (char=? #\: (string-ref str i)))
	     (cond ((= (+ 1 i) len) #f)
		   ((>= i len) name)
		   (else (let ((nstr (substring str (+ 1 i) len)))
			   (or (string->number nstr)
			       (string->symbol nstr))))))))
      name))

(define (negated? s)
  (set! s (symbol->string s))
  (and (> (string-length s) 1)
       (memv (string-ref s (+ -1 (string-length s))) '(#\# #\- #\_))
       #t))

;;;			      -=-=-=-=-

;;; Time-stamp: is an EMACS supported date and author field.
(define time-stamp:time #f)
(define time-stamp:author "M. Hatter")
(define (Time-stamp: string)
  (define cc_time (make-vector 9 0))
  (require 'posix-time)
  (require 'scanf)
  (cond ((eqv? 7 (sscanf string "%4d%*[-/]%2d%*[-/]%2d %2d:%2d:%2d %s"
			 (vector-ref cc_time 5)
			 (vector-ref cc_time 4)
			 (vector-ref cc_time 3)
			 (vector-ref cc_time 2)
			 (vector-ref cc_time 1)
			 (vector-ref cc_time 0)
			 time-stamp:author))
	 (vector-set! cc_time 8 -1)
	 (vector-set! cc_time 4 (+ -1 (vector-ref cc_time 4)))
	 (if (> (vector-ref cc_time 5) 1900)
	     (vector-set! cc_time 5 (+ -1900 (vector-ref cc_time 5))))
	 (set! time-stamp:time (mktime cc_time)))
	(else (set! time-stamp:time #f)
	      (set! time-stamp:author #f))))
;;; The rest are *NOT* Emacs supported fields.
(define time-stamp:revision 1)
(define (Revision-stamp: obj)
  (set! time-stamp:revision obj))
(define time-stamp:company "Hyperactive Engineering")
(define (Company-stamp: obj)
  (set! time-stamp:company obj))
(define time-stamp:configuration 1)
(define (Configuration-stamp: obj)
  (set! time-stamp:configuration obj))

(define cvs:time #f)
(define cvs:author "W. Rabbit")
(define cvs:filename #f)
(define cvs:version #f)
(define (cvs-id: string)
  (require 'posix-time)
  (require 'scanf)
  (require 'string-search)
  (let ((keyword #f)
	(filename #f)
	(version #f)
	(year #f)
	(month #f)
	(day #f)
	(hour #f)
	(minute #f)
	(second #f)
	(author #f))
    (let* ((i (string-index string #\$))
	   (n (and i
		   (sscanf (substring string i (string-length string))
			   "$%[^:]: %s %s %d/%d/%d %d:%d:%d %s"
			   keyword filename version year month day hour
			   minute second author))))
      (cond ((and (eqv? n 10) (member keyword '("Header" "Id")))
	     (set! cvs:filename filename)
	     (set! cvs:version version)
	     (set! cvs:author author)
	     (set! cvs:time
		   (mktime (vector second minute hour day (- month 1)
				   (if (> year 1900) (- year 1900) year)
				   0 0 -1))))
	    (else
	     (synch:warn 'cvs-id: "unrecognized format" string))))))
