;;;; "models.scm" utility models for synchronous simulation.	-*-scheme-*-
;;; Copyright (C) 1997, 1998 Aubrey Jaffer

;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

(define (make-ram-array prototype . bounds)
  (let ((ara (apply make-uniform-array prototype bounds)))
    (array-fill! ara 0)
    ara))

(define (make-ram length . prototype)
  (set! prototype (if (null? prototype) #f (car prototype)))
  (let ((ara (make-uniform-array prototype length)))
    (uniform-vector-fill! ara 0)
    ara))

(define (make-fifo length . prototype)
  (set! prototype (if (null? prototype) #f (car prototype)))
  (let ((ara (make-uniform-array prototype (list -2 (+ -1 length)))))
    (uniform-vector-fill! ara 0)
    ara))
(define (fifo:clear fifo)
  (uniform-vector-fill! fifo 0)
  fifo)
(define (fifo:first fifo)
  (logaref fifo (modulo (logaref fifo -2)
			  (+ 1 (cadar (array-shape fifo))))))
(define (fifo:remove-first fifo)
  (if (fifo:empty? fifo) (synch:error 'fifo-underflow fifo))
  (array-set fifo
	     (modulo (+ 1 (logaref fifo -2))
		     (* 2 (+ 1 (cadar (array-shape fifo)))))
	     -2))
(define (fifo:insert-last fifo datum)
  (if (fifo:full? fifo) (synch:error 'fifo-overflow fifo))
  (array-set (array-set fifo datum (modulo (logaref fifo -1)
					   (+ 1 (cadar (array-shape fifo)))))
	     (modulo (+ 1 (logaref fifo -1))
		     (* 2 (+ 1 (cadar (array-shape fifo)))))
	     -1))
(define (fifo:empty? fifo)
  (= (logaref fifo -1) (logaref fifo -2)))
(define (fifo:full? fifo)
  (define len (+ 1 (cadar (array-shape fifo))))
  (= len (modulo (- (logaref fifo -1) (logaref fifo -2))
		 (* 2 len))))
(define (fifo:fullness fifo)
  (modulo (- (logaref fifo -1) (logaref fifo -2))
	  (* 2 (+ 1 (cadar (array-shape fifo))))))

(define (bounds->prototype bounds)
  (if bounds
      (+ -1 (ash 1 (abs (- (cadr bounds) (caddr bounds)))))
      #f))
(define (must-be-id name)
  (if (not (and (symbol? name)
		(ptag-prefix name)))
      (slib:warn 'synch:fifo name 'not 'id)))

(defmacro synch:fifo (name length data-source write-strobe read-strobe
			   data-first fullness empty? full?)
  (must-be-id name)
  (and empty? (must-be-id empty?))
  (and full? (must-be-id full?))
  (let* ((bounds (vector-bounds name))
	 (name-no-pre (ptag-stripped name))
	 (pre (symbol-append (ptag-prefix name) ":"))
	 (nam (ptag-stripped (if bounds (car bounds) name)))
	 (idxbits (if (number? length)
		      (symbol-append "[" (- (integer-length length) 1)
				     "..0]")
		      "[31..0]"))
	 (outidx (symbol-append nam '-outidx idxbits))
	 (inidx (symbol-append nam '-inidx idxbits)))
    (if (not fullness) (set! fullness (symbol-append
				       (if bounds (car bounds) name)
				       '-fullness)))
    (must-be-id fullness)
    ;;(if (vector-bounds first) (slib:warn 'synch:fifo first 'should-not-have-bounds))
    `(begin
       (synch:define 'pre ,name-no-pre
		     (make-ram ,length ,(bounds->prototype bounds))
		     #f #f
		     (if ,write-strobe
			 (array-set ,name-no-pre
				    ,data-source
				    (modulo ,inidx ,length))
			 ,name-no-pre))
       (synch:define ,pre ,outidx 0 #f #f
		     (if ,read-strobe
			 (modulo (+ 1 ,outidx) (* 2 ,length))
			 ,outidx))
       (synch:define ,pre ,inidx 0 #f #f
		     (if ,write-strobe
			 (modulo (+ 1 ,inidx) (* 2 ,length))
			 ,inidx))
       (synch:defmacro ,fullness
		       (modulo (- ,inidx ,outidx) (* 2 ,length)))
       ,(if full?
	    `(synch:defmacro ,full? (= ,length ,(ptag-stripped fullness)))
	    #f)
       ,(if empty?
	    `(synch:defmacro ,empty? (zero? ,(ptag-stripped fullness)))
	    #f)
       (synch:defmacro ,data-first
		       (logaref ,name-no-pre (modulo ,outidx ,length))))))
