;;;; "simsynch.scm" simulate synchronous systems.	-*-scheme-*-
;;; Copyright (C) 1995, 1996, 1997 Aubrey Jaffer

;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

;; $Header: /usr/local/cvsroot/synch/simsynch.scm,v 1.30 1999/01/07 20:34:39 jaffer Exp $

(display ";			 loading SIMSYNCH") (newline)
(require 'database-browse)
(define b browse)

;;(provide 'check-turnaround)

(require 'sort)
(require 'line-i/o)
(require 'database-utilities)
(require 'common-list-functions)
(require 'logic)			;(in-vicinity (program-vicinity) "logic.scm")

(define (create-board board-name dbtype)
  (define boardb (create-database (string-append
				   (symbol->string board-name) ".db")
				  dbtype))

  (define-tables boardb
    '(constants
      ((name symbol))
      ((value expression))
      ())
    '(pld
      ((name string))
      ((package string))
      (("MACH355" "144PQFP")
       ("MACH445" "100PQFP")
       ("MACH465" "208PQFP")
       ("PALCE610" "24DIP300")
       ("OR3PS208" "208SQFP2")
       ("OR3PS240" "240SQFP2")
       ("OR3BA256" "256PBGA")
       ("OR3BA352" "352PBGA")
       ("OR3BC432" "432EBGA")
       ("OR3BC600" "600EBGA")
       ("*"	"?")
       (""	"?")))

    '(ptag
      ((name symbol))
      ((pld-type pld)
       (signature expression (lambda (x) (or (not x) (number? x)))))
      ((int "" #f)
       (hid "" #f)
       (* "*" #f)))

    '(pin-type
      ((name atom))
      ()
      ((output)
       (input/output)
       (wire-and)
       (wire-or)
       (clock)
       (input)
       (unused)
       (internal)))

    '(pin
      ((ptag ptag)
       (name symbol))
      ((pin-type pin-type)
       (pin-name expression)		;'?, number, symbol, or list of same.
       (output-enable expression)
       (paired-input expression (lambda (x) (or (not x) (symbol? x)))))
      ())

    '(pindex
      ((ptag ptag)
       (pin-name atom))
      ((symbol-name symbol)
       (pin-type pin-type)
       (signal-name expression))
      ())

    '(block
      ((name symbol))
      ((clock symbol)			;really signal foreign-key
       (reset symbol)			;really signal foreign-key
       )
      ())

    '(check
      ((block block)
       (order uint))
      ((expression expression))
      ())

    '(signal-type
      ((name symbol))
      ()
      ((registered)
       (registered-input)
       (combinatorial)
       (hidden)				;used for inputs
       (macro)))

    '(signal
      ((block block)
       (name symbol))
      ((defining-ptag ptag)
       (signal-type signal-type)
       (initial-state expression)
       (next-function expression)
       (order uint))
      ())

    '(used-in
      ((name symbol)
       (block block))
      ()
      ())
    )
  boardb)

(define *block* #f)
(defvar *design* 'test)
(define *board* (create-board *design* 'alist-table))

(define (synch:register-block block . args)
  (let* ((clock (if (null? args)
		    (symbol-append block '- 'clock)
		    (car args)))
	 (reset (if (or (null? args) (null? (cdr args)))
		    (symbol-append block '- 'reset-)
		    (cadr args))))
    ((((*board* 'open-table) 'block #t) 'row:insert)
     (list block clock reset))
    (set! *block* block)))

(define (synch:register-ptag tag ptag . args)
  (define signature (if (null? args) #f (car args)))
  ((((*board* 'open-table) 'ptag #t) 'row:insert)
   (list tag ptag signature)))

(define ptag-table ((*board* 'open-table) 'ptag #t))
(define pld->package (((*board* 'open-table) 'pld #f) 'get 'package))
(define ptag->pld (ptag-table 'get 'pld-type))
(define ptag->sig (ptag-table 'get 'signature))
(define (ptag->package ptag) (pld->package (ptag->pld ptag)))

(define pin-table ((*board* 'open-table) 'pin #t))
(define pin-table:row-insert (pin-table 'row:insert))
(define pin-table:row-retrieve (pin-table 'row:retrieve))

(define ordering:counter 100)
(define signal-table ((*board* 'open-table) 'signal #t))
(define signal-table:row-insert (signal-table 'row:insert))
(define signal-table:row-remove (signal-table 'row:remove))
(define signal-table:row-retrieve (signal-table 'row:retrieve))
(define signal-table:get-init-fun (signal-table 'get 'initial-state))
(define signal-table:get-next-fun (signal-table 'get 'next-function))

(define used-in-table ((*board* 'open-table) 'used-in #t))
(define used-in-table:row-insert (used-in-table 'row:insert))
;;(define used-in-table:row-remove (used-in-table 'row:remove))
(define used-in-table:row-retrieve* (used-in-table 'row:retrieve*))
(define used-in-table:get-block (used-in-table 'get 'block))

(define block-table ((*board* 'open-table) 'block #f))

(define (mark-uses block next-function)
  (cond ((symbol? next-function)
	 (if (not (used-in-table:get-block next-function block))
	     (used-in-table:row-insert (list next-function block))))
	((and (list? next-function) (pair? next-function))
	 (for-each (lambda (expr) (mark-uses block expr))
		   next-function))))

(define (used-by-other-blocks? signal block)
  (not (null? (used-in-table:row-retrieve*
	       signal (lambda (key) (not (eq? key block)))))))

(define constant-table ((*board* 'open-table) 'constants #t))
(define constant-table:insert
  (let ((inserter (constant-table 'row:insert)))
    (lambda args
      (inserter args))))
(define constant-table:row-retrieve (constant-table 'row:retrieve))

(defmacro defconst (name value)
  (cond ((list? name) `(defconst ,(car name) (lambda ,(cdr name) ,value)))
	(else (cond ((not (slib:eval `(defined? ,name))))
		    ((and (symbol? name) (eqv? (slib:eval value)
					       (slib:eval name))))
		    (else (slib:error 'trying-to-defconst name
				      'to-different-value value)))
	      `(begin (define ,name ,value)
		      (constant-table:insert ',name ,name)))))
;;; Eval constant
(define (ceval thing)
  (cond ((number? thing) thing)
	((not (symbol? thing)) thing)
	((constant-table:row-retrieve thing) =>
	 cadr)
	(else thing)))

(define (for-each-ptag proc)
  ((ptag-table 'for-each-row)
   (lambda (ptag-row)
     (define str (cadr ptag-row))
     (if (and (not (equal? "" str))
	      (or (char-alphabetic? (string-ref str 0))
		  (char-numeric? (string-ref str 0))))
	 (apply proc ptag-row)))))

(define check-table ((*board* 'open-table) 'check #t))
(define synch:insert-check
  (let ((check-table:row-insert (check-table 'row:insert)))
    (lambda (check)
      (set! ordering:counter (+ ordering:counter 1))
      (check-table:row-insert (list *block* ordering:counter check)))))

;;(SYNCH:DEFINE-PIN
;; ptag name pin-type pin-name output-enable paired-input)
;;(SYNCH:DEFINE-SIGNAL
;; [block] name defining-ptag signal-type initial-state next-function)
(define (synch:define-pin ptag . args)
  (cond (ptag
	 (if (not ((ptag-table 'row:retrieve) ptag))
	     (synch:register-ptag ptag ""))
	 (pin-table:row-insert (cons ptag args))
	 (apply
	  (lambda (name pin-type pin-name output-enable paired-input)
	    (cond
	     ((eq? pin-name '?))
	     ((and (or (list? pin-name) (symbol? pin-name))
		   (vector-bounds (or paired-input name)))
	      => (lambda (bounds)
		   (let* ((first (cadr bounds))
			  (last (caddr bounds))
			  (range (- last first))
			  (incer (sign range)))
		     (do ((i 0 (+ incer i)))
			 ((positive? (* incer (- i range))))
		       (pindex:row-insert
			(list ptag
			      (if (pair? pin-name)
				  (list-ref pin-name (abs i))
				  (symbol-append pin-name (abs i)))
			      '?
			      pin-type
			      (refvectorize (or paired-input name) (+ i first))
			      ))))))
	     ((or (and (number? pin-name) (not (zero? pin-name)) pin-name)
		  (and (symbol? pin-name) (ptag-stripped pin-name)))
	      => (lambda (pname)
		   (pindex:row-insert
		    (list ptag pname '? pin-type name))))))
	  args))))
(define (synch:define-signal name ptag . args)
  (set! ordering:counter (+ ordering:counter 1))
  (cond (ptag (if (not ((ptag-table 'row:retrieve) ptag))
		  (synch:register-ptag ptag ""))))
  (signal-table:row-insert (append (list *block* name (or ptag '*))
				   args
				   (list ordering:counter))))

(define (synch:define-bus name float-limit . connections)
  (define name-drv (symbol-append name '- 'drv))
  #+check-turnaround
  (define name-lst (symbol-append name '- 'lst))
  (if (eqv? 1 (length connections))
      (set! connections (car connections)))
  (synch:define-signal name 'hid 'macro #f
		       `(cond ,@connections (else ,(- -1 float-limit))))
  #+check-turnaround
  (synch:define-signal
   name-lst 'hid 'registered (- -1 float-limit) name)
  (synch:define-signal
   name-drv 'hid 'registered (- -1 float-limit)
   `(case (count ,@(remove #f (map car connections)))
      ((0) (cond ((eqv? -1 ,name-drv)
		  (synch:warn ',name 'floating ,float-limit 'cycles)
		  -1)
		 ((and (number? ,name-drv) (negative? ,name-drv))
		  (+ 1 ,name-drv))
		 (else ,(- -1 float-limit))))
      ((1) (let ((idx (cond ,@(let ((idx -1))
				(map (lambda (con)
				       (set! idx (+ 1 idx))
				       `(,(car con) ,idx))
				     connections))
			    (else #f))))
	     #+check-turnaround
	     (cond ((eqv? idx ,name-drv))
		   ((negative? ,name-drv))
		   ((negative? idx))
		   ((eqv? ,name ,name-lst))
		   (else (synch:warn ',name 'no-turnaround-time ,name-drv idx)))
	     idx))
      ,@(if (some boolean? (map car connections))
	    `(((2) ,name-drv))
	    '())
      (else (apply synch:error ',name 'bus-driving-conflict
		   (list ,@(map car connections)))
	    (cond ,@(let ((idx -1))
		      (map (lambda (con)
			     (set! idx (+ 1 idx))
			     `(,(car con) ,idx))
			   connections))
		  (else #f))))))

(defmacro synch:defcheck (check)
  (tprint 'synch:defcheck check)
  (synch:insert-check check)
  "check")

;;; Synch:pre is a macro that saves typing.

(defmacro synch:pre (signal)
  (tprint 'synch:pre signal)
  (let ((stepper (signal-table:get-next-fun *block* signal)))
    (case stepper
      ((#f) (synch:error "signal not known:" *block* signal))
      ((?) (synch:error "signal not yet set:" *block* signal))
      (else stepper))))

(defmacro synch:defmacro (name expression)
  (define ptag (ptag-prefix name))
  (set! name (ptag-stripped name))
  (tprint 'synch:defmacro ptag name)
  (synch:define-signal name ptag 'macro #f expression)
  (symbol->string name))

(defmacro synch:defshare (pin-nam name)
  (define ptag (ptag-prefix name))
  (set! name (ptag-stripped name))
  (tprint 'synch:defshare ptag name)
  (cond ((signal-table:row-retrieve *block* name)
	 (synch:define-pin ptag name 'input (ptag-stripped pin-nam) #f name))
	(else (synch:error "shared signal not defined:" name)))
  (symbol->string name))

(defmacro synch:definput (pin-nam name expression)
  (define ptag (or (ptag-prefix name) (ptag-prefix pin-nam)))
  (set! name (ptag-stripped name))
  (tprint 'synch:definput ptag name)
  (synch:define-signal name ptag 'hidden #f expression)
  (synch:define-pin ptag name 'input (ptag-stripped pin-nam) #f name)
  (symbol->string name))

(defmacro synch:defreginput (pin-nam fullname expression)
  (let* ((ptag (or (ptag-prefix fullname) (ptag-prefix pin-nam)))
	 (name (ptag-stripped fullname))
	 (fullprename (symbol-append ptag ': 'pre: name)))
    (tprint 'synch:defreginput ptag name)
    `(begin
       (define ,name #f)		;This BAD Hack is spreading!
					;Needed for self reset test in scm2mach
					;registered inputs clear to #f
       (synch:definput ,(ptag-stripped pin-nam) ,fullprename ,expression)
       (synch:define 0 ,fullname ,name #f #f)
       (synch:set! ,name ,(symbol-append 'pre: name)))))

(defmacro synch:define (pin-nam signal init-state pin oen . opt-next)
  (define ptag (or (ptag-prefix signal) (ptag-prefix pin-nam)))
  (set! opt-next (if (null? opt-next) '? (car opt-next)))
  (tprint 'synch:define ptag signal)
  (cond (ptag
	 (set! signal (ptag-stripped signal))
	 (synch:define-pin ptag signal
			   (case oen
			     ((#f) 'internal)
			     ((#t) 'output)
			     (else 'input/output))
			   (ptag-stripped pin-nam)
			   oen
			   (if oen (or pin signal) pin))))
  (synch:define-signal signal ptag 'registered init-state opt-next)
  (symbol->string signal))

(defmacro synch:set/reset! (signal turn-on turn-off)
  (tprint 'synch:set/reset! signal)
  (synch:insert-check
   `(cond ((and ,turn-on ,turn-off)
	   (synch:warn ',signal 'being-turned-on-and_off_))))
  `(synch:set! ,signal (and (not ,turn-off) (or ,turn-on ,signal))))

(defmacro synch:set! (signal new-state)
  (tprint 'synch:set! signal)
  (let* ((row (signal-table:row-remove *block* signal))
	 (last2 (last row 2)))
    (cond ((not row) (synch:error "signal not known:" signal))
	  ((eq? '? (car last2))
	   (signal-table:row-insert
	    (append (butlast row 2) (cons new-state (cdr last2)))))
	  ((and (pair? (car last2)) (eq? 'compose-sets (caar last2)))
	   (signal-table:row-insert
	    (append (butlast row 2)
		    (cons (append (car last2) `((lambda (,signal) ,new-state)))
			  (cdr last2)))))
	  (else
	   (signal-table:row-insert
	    (append (butlast row 2)
		    (cons `(compose-sets ,signal
					 (lambda (,signal) ,(car last2))
					 (lambda (,signal) ,new-state))
			  (cdr last2))))))
    (symbol->string signal)))

(define synch:step '())
(define synch:status #f)
(defmacro define-synchronous-system (system-name . body)
  (cond ((not (eq? system-name *block*))
	 (slib:error "defining system-name different from *block*?" system-name
		     *block*)))
  (synch:define-signal 'synch:prin-count 'INT: 'registered -1 'synch:prin-count)
  (synch:define-signal 'synch:count 'INT: 'registered -1 '(+ 1 synch:count))
  (let ((names '())
	(inits '())
	(macros '())
	(hiddens '())
	(steppers '())
	(block->reset (((*board* 'open-table) 'block #f) 'get 'reset))
	(synch:blocks ((((*board* 'open-table) 'block #f) 'get* 'name))))
    (for-each
     (lambda (sig-row)
       (apply
	(lambda (block name defining-ptag signal-type
		       initial-state next-function order)
	  (case signal-type
	    ((registered combinatorial macro)
	     (mark-uses block next-function)))
	  (case signal-type
	    ((registered) (if (eq? '? next-function)
			      (synch:error "signal not set:" sig-row))
	     (set! names (cons name names))
	     (set! inits (cons initial-state inits))
	     (set! steppers (cons next-function steppers)))
	    ((hidden) (set! hiddens (cons (list name next-function) hiddens)))
	    ((macro combinatorial)
	     (set! macros (cons (list name next-function) macros)))
	    (else (synch:error "unknown signal-type" sig-row))))
	sig-row))
     (sort ((signal-table 'row:retrieve*) *block*)
	   (lambda (row1 row2) (> (list-ref row1 6) (list-ref row2 6)))))
    ((*board* 'write-database) (string-append (symbol->string *design*) ".db"))
    (set! synch:step
	  (slib:eval
	   `(lambda ()
	      (or?
	       ,@(map (lambda (sys-name)
			(define reset-name (block->reset sys-name))
			(define state-name (symbol-append sys-name ':state))
			`(cond (,(if (negated? reset-name)
				     `(not ,reset-name)
				     reset-name)
				(set! ,state-name #f)
				#f)
			       ((not ,state-name)
				(set! ,state-name
				      (slib:eval
				       (cons 'list
					     ,(symbol-append
					       sys-name ':state-initializers))))
				#t)
			       ((positive? ,(symbol-append sys-name ':queue))
				(set! ,(symbol-append sys-name ':queue)
				      (+ -1 ,(symbol-append sys-name ':queue)))
				(set! ,state-name
				      (@apply
				       ,(symbol-append sys-name ':state-step)
				       ,state-name))
				#t)
			       (else #f)))
		      synch:blocks)))))	;(reverse )
    (set! synch:status
	  (slib:eval
	   `(lambda ()
	      ;;(display "...")
	      (if synch:stop (printf "STOPPED!"))
	      ,@(map (lambda (block)
		       (let ((reset-name (block->reset block)))
			 `(apply printf " %s: %s q=%d r=%d p=%d"
				 ',(string-upcase (symbol->string block))
				 (if ,(if (negated? reset-name) reset-name
					  `(not ,reset-name))
				     "run"
				     "rst")
				 ,(symbol-append block ':queue)
				 (reverse
				  (last (or ,(symbol-append block ':state)
					    (list 0 0))
					2)))))
		     synch:blocks)	;(reverse )
	      (newline))))
    `(begin
       (define ,(symbol-append system-name ':queue) 0)
       (define (,(symbol-append system-name '-sim) num)
	 (set! ,(symbol-append system-name ':queue)
	       (max num ,(symbol-append system-name ':queue)))
	 (kb:announce-polling 'break)
	 (simulate!))
       (define ,(symbol-append system-name ':state) #f)
       (define ,(symbol-append system-name ':state-initializers) ',inits)
       (defmacro ,system-name (name)
	 (if (and (list? name) (pair? name) (eq? 'quote (car name)))
	     (set! name (cadr name)))
	 (let ((vals (symbol-append ',system-name ':state))
	       (idx (position name ',names)))
	   (if (not idx)
	       (slib:error name 'not-registered-in vals)
	       `(if ,vals
		    (list-ref ,vals ,idx)
		    (slib:eval
		     (list-ref
		      ,',(symbol-append system-name ':state-initializers)
		      ,idx))))))
       (define (,(symbol-append system-name ':state-step) ,@names)
	 (fluid-let ((*block* ',*block*))
	   (let* ,macros
	     (let* ,hiddens
	       ,@(map caddr
		      (sort ((check-table 'row:retrieve*) *block*)
			    (lambda (row1 row2) (> (cadr row1) (cadr row2)))))
	       ,@body
	       (let ((new-state (list ,@steppers)))
		 (update-deferred-sets!)
		 new-state))))))))

(define simulating? #f)
(define (simulate!)
  (cond (simulating?)
	(synch:stop (synch:status) (kb:break))
	((kb:break-on-hit))
	(else (set! simulating? #t)
	      (let ((more-to-do? (synch:step)))
		(set! simulating? #f)
		(if more-to-do? (simulate!))))))

(define synch:stop #f)
(define (synch:reset) (set! synch:stop #f))

;;; Print out as many warnings and errors as possible before stopping.
(require 'object->string)
(defvar print-timing #t)
(defvar print-1-timing #f)
(define saved-print-timing #f)
(defvar print-chat #t)
(define synch:msg-list '())
(defvar talkative #t)
(define (tprint . args)
  ;;(if talkative (apply print args))
  #f)

(defvar report slib:error)

(define (synch:chat msg)
  (set! synch:msg-list (cons msg synch:msg-list))
  #t)

(define (synch:warn . args)
  (define cep (current-error-port))
  (cond
   ((printf:format? (car args))
    (apply fprintf cep args))
   (else
    (display "WARN: " cep)
    (if (not (null? args))
	(begin (display (car args) cep)
	       (for-each (lambda (x) (display #\  cep) (write x cep))
			 (cdr args))))))
  (newline cep)
  (force-output cep)
  (set! print-1-timing #t))

(define (printf:format? str)
  (cond ((not (string? str)) #f)
	 ((< 100 (string-length str)) #f)
	 (else (memv #\% (string->list str)))))

(define (synch:error . args)
  (define cep (current-error-port))
  (cond
   ((printf:format? (car args))
    (apply fprintf cep args))
   (else
    (display "ERROR: " cep)
    (if (not (null? args))
	(begin (display (car args) cep)
	       (for-each (lambda (x) (display #\  cep) (write x cep))
			 (cdr args))))))
  (newline cep)
  (force-output cep)
  (set! print-1-timing #t)
  (set! synch:stop #t))

(defmacro synch:prin-header (h cnt)
  (cond ((and (list? h) (pair? h) (eq? (car h) 'quote))
	 (set! h (cadr h))))
  (let* ((ss (map (lambda (i)
		    (cond ((string? i) (string-upcase! i))
			  ((char? i) "")
			  (else (object->string i))))
		  h))
	 (sl (map string-length ss))
	 (ml (apply max sl))
	 (ll (map (lambda (s l)
		    (string->list
		     (if (zero? l) (make-string ml #\nul)
			 (string-append s (make-string (- ml l) #\ )))))
		  ss sl))
	 (rl (apply map list ll)))
    `(let ()
       (dprintf "%4d\\n" ,cnt)
       ,@(map
	  (lambda (l) `(dprintf ,l))
	  (map (lambda (l)
		 (string-append
		  (apply
		   string-append
		   " "
		   (if (char=? #\nul (car l)) "" (string (car l)))
		   (map (lambda (c)
			  (string-append
			   "  " (if (char=? #\nul c) "" (string c))))
			(cdr l)))
		  ;;(cond (cnt (display "  ")(display cnt) (set! cnt #f)))
		  "\\n"))
	       rl))
       "synch:prin-header")))

(define (synch:prin1 x)
  (case x
    ((#t) (display "  |"))
    ((#f) (display " | "))
    (else
     (display #\ )
     (cond ((char? x) (display x))
	   ((number? x)
	    (let ((s (number->string x 16))
		  (d 2))
	      (do ((l (string-length s) (+ 1 l)))
		  ((>= l d) (display s))
		(display #\0))))
	   ((string? x) (display x))
	   (else (write x))))))

(define synch:prin-args #f)
(define synch:last-printed-count -1)

(defmacro synch:print args
  (define names '())
  (define argl '())
  (do ((l args (cdr l)))
      ((null? l))
    (cond ((string? (car l))
	   (SET! names (cons (car l) names))
	   (set! argl (cons (cadr l) argl))
	   (set! l (cdr l)))
	  ((char? (car l))
	   (set! names (cons (car l) names))
	   (set! argl (cons (car l) argl)))
	  (else
	   (set! names (cons (object->string (car l)) names))
	   (set! argl (cons (car l) argl)))))
  `(begin
     (cond
      ((or PRINT-TIMING print-1-timing)
       (set! print-1-timing #f)
       (let ((args (list ,@(reverse argl))))
	 (cond ((not (eqv? synch:last-printed-count (+ -1 synch:count)))
		(synch:status)
		(if (not (zero? (dma:dword-read-count io:brd)))
		    (printf "%d dwords read\n"
			    (dma:dword-read-count io:brd)))))
	 (cond ((or (not (equal? args synch:prin-args))
		    (not (null? synch:msg-list)))	; (and PRINT-CHAT)
		(set! synch:prin-count (+ 1 synch:prin-count))
		(set! synch:prin-args args)
		(cond ((zero? (remainder synch:prin-count 50))
		       (synch:prin-header ',(reverse names) synch:count)
		       (newline)))
		(for-each synch:prin1 args)
		(for-each (lambda (m) (display #\ ) (display m))
			  synch:msg-list)
		(newline))))
       (set! synch:last-printed-count synch:count))
      ((and PRINT-CHAT (or ;;(zero? (remainder synch:count 50))
			   (not (null? synch:msg-list))))
       (printf "%5.5d %4.4d" synch:count (dma:dword-read-count io:brd))
       (for-each (lambda (m) (display #\ ) (display m)) synch:msg-list)
       (newline)))
     (set! synch:msg-list '())))

;   ================================================================

(define pindex ((*board* 'open-table) 'pindex #t))
(define pindex:row-insert (pindex 'row:insert))
(define pindex:row-retrieve (pindex 'row:retrieve))

(define (expand-range pair)
  (if (list? pair) pair
      (do ((l '() (cons j l))
	   (inc (if (> (cdr pair) (car pair)) -1 +1))
	   (j (cdr pair) (+ inc j)))
	  ((= j (car pair)) (cons j l)))))

(define (db:read-pins ptag name)
  (call-with-input-file (string-append name ".pin")
    (lambda (iport)
      (let ((shape (read iport)))
	(cond
	 ((not (list? shape)) (slib:error "not valid format" shape))
	 ((memq (car shape) '(pqfp plcc lcc soj))
	  (set-car! (cdr shape) (expand-range (cadr shape)))
	  (do ((rows (cadr shape) (cdr rows)))
	      ((null? rows))
	    (let* ((pnum (read iport))
		   (pnam (read iport))
		   (ptyp (read iport)))
	      (cond ((not (equal? pnum (car rows)))
		     (slib:error "pins out of sequence"
				 (car rows) "~=" pnum)))
	      (pindex:row-insert
	       (list ptag
		     pnum
		     pnam
		     (cdr (or (assq ptyp
				    '((out . output)
				      (i/o . input/output)
				      (hiZ . input/output)
				      ;;(wire-and)
				      ;;(wire-or)
				      (clk . clock)
				      (in . input)
				      (v . unused)
				      ;;(internal)
				      ))
			      (cons ptyp ptyp)))
		     #f)))))
	 (else (slib:error "unknown format" shape)))))))

;   ================================================================

;;(trace synch:define-signal synch:define-pin synch:register-block)
;;(ta "simsync.scm")

;; Output binary format.

(define (dump:spew . args)
  (define len (length args))
  (display #\@ pf:dumpport)
  (display (number->string dump:address 16) pf:dumpport)
  (set! dump:address (+ dump:address len))
  (map (lambda (x)
	 (display #\space pf:dumpport)
	 (display (case x
		    ((#t) #\1)
		    ((#f) #\0)
		    (else (cond ((not (number? x))
				 (report 'dump:spew funny-value x) #\z)
				((negative? x) #\z)
				(else (number->string x 16)))))
		  pf:dumpport))
       args)
  (newline pf:dumpport))
